/*jslint node: true*/
'use strict';

var meta = require('./package.json'),
    gutil = require('gulp-util'),
    through = require('through2'),
    fs = require('fs'),
    htmlparser = require("htmlparser2"),
    error = false,
    errorBreak = false;

/** 
 * Validation Error Printout
 * abort = true => breaks build if flag is set
 */
function validationError(tag, mistake, guideline, abort) {

    if (typeof mistake === 'string') {
        mistake = mistake.substr(mistake.search(/\S|$/));

        if (mistake.length > 60) {
            mistake = mistake.substr(0, 60) + "...";
        } else if (mistake.length === 0) {
            mistake = '[whitespace only]';
        }
    }

    error = true;

    if (abort) {
        gutil.log('[' + guideline + '] ' + gutil.colors.white(tag + ": ") + gutil.colors.red(mistake) + " [Level Red]");
        errorBreak = true;
    } else {
        gutil.log('[' + guideline + '] ' + gutil.colors.white(tag + ": ") + gutil.colors.yellow(mistake) + " [Level Orange]");
    }
}

function isLowerCase(strings) {

    var i = 0,
        character = '',
        lowercase = true;

    while (i <= strings.length) {
        character = strings.charAt(i);
        if (character !== character.toLowerCase()) {
            lowercase = false;
        }
        i = i + 1;
    }
    return lowercase;
}


function bodyStylesheet(file) {

    var style = file.match(/<body[\S\s]*<style[\S\s]*<\/style>/g),
        stylesheet = file.match(/<body[\S\s]*<link[\S\s]*rel="stylesheet"/g);

    if (style && style.length > 0) {
        validationError("Frontend performance", 'Never use CSS inline styles in the <body> section', 'CI-FE-1009', false);
    }
    if (stylesheet && stylesheet.length > 0) {
        validationError("Frontend performance", 'Never link CSS stylesheets in the <body> section', 'CI-FE-1009', false);
    }
}

function checkRemoteProtocol(name, sourcelink) {

    var r = new RegExp('^(?:[a-z]+:)?//', 'i');
    if (sourcelink) {

        if (name !== 'a' && r.test(sourcelink) && !sourcelink.includes('RB.AdaptiveImage')) {

            if (sourcelink.indexOf('//') !== 0) {
                validationError('Omit the protocol from embedded resources', '<' + name + '>: ' + sourcelink.substr(0, 30) + '...', 'CI-FE-2008', false);
            } else {
                validationError('External resources - Do not load resources from remote servers', '<' + name + '>: ' + sourcelink.substr(0, 30) + '...', 'CI-FE-1018a', false);
            }
        }
    }
}

/**
 * Check Functions via HTML parser:
 * - Images: Existing & non-empty alt attribute
 * - Elements & Attributes have to be lower case
 * - Check for language attribute in html tag
 * - script & css includes without type attribute
 * return true if error
 */
function parseHTML(file, checkremote) {

    var parser = new htmlparser.Parser({
        onopentag: function (name, attribs) {

            if (checkremote) {
                checkRemoteProtocol(name, attribs.src || attribs.href || attribs.srcset);
            }

            if (name === "html" && !attribs.lang) {
                validationError('Language attribute is missing', '<' + name + '>', 'CI-FE-2003', true);
            }
            // Img alt
            else if (name === "img" && !attribs.alt) {
                validationError("Missing or empty 'alt' tag", '<' + name + ' src="' + attribs.src + '">', 'CI-FE-2006', true);
            }
            // Iframes
            else if (name === "iframe") {
                validationError('Iframes should not be used', '<' + name + '>', 'CI-FE-2010', false);
            }
            if (!isLowerCase(name)) {
                validationError('Markup must always be written in lowercase', '<' + name + '>', 'CI-FE-2001', true);
            }

            if (attribs.type === "text/javascript") {
                validationError('JavaScript includes - Remove type="text/javascript"', '<' + name + ' type="' + attribs.type + '" ...>', 'CI-FE-2007', true);
            } else if (attribs.type === "text/css") {
                validationError('CSS includes - Remove type="text/css"', '<' + name + ' type="' + attribs.type + '" ...>', 'CI-FE-2007', true);
            }

        },
        onattribute: function (name, value) {

            if (!isLowerCase(name)) {
                validationError("Markup must always be written in lowercase", 'Attribute "' + name + '"', 'CI-FE-2001', false);
            }
        }
    }, {
        decodeEntities: true,
        lowerCaseTags: false,
        lowerCaseAttributeNames: false
    });
    parser.write(file);
    parser.end();
}

/**
 * Check for Viewport tag existence (if option set)
 * return true if error
 */
function checkViewport(file) {

    if (file.includes('<meta') || file.includes('<head>')) {

        var found = false,
            parser = new htmlparser.Parser({
                onopentag: function (name, attribs) {

                    if (name === "meta" && attribs.name === "viewport") {

                        found = true;
                        var nospace = attribs.content.replace(/\s/g, "");

                        if (!nospace.includes('user-scalable=no') || !nospace.includes('width=device-width') || !nospace.includes('initial-scale=1') || !nospace.includes('maximum-scale=1')) {
                            validationError('Viewport settings - Viewport tag is incorrect', attribs.content, 'CI-FE-2004b', false);
                        }
                    }
                }
            }, {
                decodeEntities: true
            });
        parser.write(file);
        parser.end();

        if (!found) {
            validationError('Viewport settings - Viewport tag is missing', '<head>', 'CI-FE-2004', false);
        }
    }
}

/**
 * Check for correct indentation: tabs / X spaces
 * return true if error
 */
function checkIndentation(file, indent, checkInTags) {

    var lines = file.split(/\r?\n/),
        pos = 0,
        area = "",
        areacheck = "",
        count = 0,
        insideTag = false;

    lines.forEach(function (line) {

        count = count + 1;

        // whitespace before actual code
        pos = line.search(/\S|$/);
        area = line.substr(0, pos);
        areacheck = "";

        if (pos > 0) {

            // replace remaining whitespace with correct indent format until no longer possible
            while (area.length >= indent.length && area !== areacheck) {
                areacheck = area;
                area = area.replace(indent, '');
            }

            // if whitespace remains -> wrong format
            if (area.length > 0 && (checkInTags || !insideTag)) {
                validationError('Wrong indentation (line ' + count + '). Use auto formatter', line, 'CI-FE-2001', false);
            }
        }

        // opening tag / closing tag?
        if ((line.match(/</g) || []).length > (line.match(/>/g) || []).length) {
            insideTag = true;
        } else if ((line.match(/</g) || []).length < (line.match(/>/g) || []).length) {
            insideTag = false;
        }

    });
}

/**
 * Check for single quotes on attributes
 * return true if found (error)
 */
function singleQuotes(file) {

    var res = file.match(/\s\w+='(\w|\s)*'/g);

    if (res && res.length > 0) {
        validationError('Always use double quotes, never single quotes, on attributes', res, 'CI-FE-2001', false);
    }
}

/**
 * Main Module
 */
module.exports = function validate(options) {

    /* Default Values - Check passed parameters */

    if (!options) {
        options = {};
    }
    if (!options.indent) {
        options.indent = '    '; // Enter x spaces or \t
    }
    if (typeof options.checkViewport !== 'boolean') {
        options.checkViewport = true;
    }
    if (typeof options.checkIndent !== 'boolean') {
        options.checkIndent = true;
    }
    if (typeof options.checkIndentInTags !== 'boolean') {
        options.checkIndentInTags = false;
    }
    if (typeof options.checkRemote !== 'boolean') {
        options.checkRemote = true;
    }

    var errorOutputList = [];

    function transformFunction(file, encoding, callback) {

        fs.readFile(file.path, 'utf8', function (err, str) {

            error = false;

            if (!str) {
                validationError('', 'File is empty or null', 'CI-FE-WARNING', false);
            } else {

                parseHTML(str, options.checkRemote);
                singleQuotes(str);
                bodyStylesheet(str);

                if (options.checkIndent) {
                    checkIndentation(str, options.indent, options.checkIndentInTags);
                }

                if (options.checkViewport) {
                    checkViewport(str);
                }

                if (str.includes('<head>') && !str.includes('charset="utf-8"') && !str.includes('charset=utf-8')) {
                    validationError('Undefined or wrong encoding', 'It should be: charset="utf-8"', 'CI-FE-2002', true);
                }
            }

            if (error) {
                gutil.log(gutil.colors.cyan("File: " + file.path));

                if (options.exitOnError && errorBreak) {
                    gutil.log(gutil.colors.red("Critical validation error -> EXIT") + " (exitOnError is active)");
                    process.exit(1);
                }
                gutil.log();
            }
        });

        return callback(null, file);
    }

    function errorOutput(callback) {
        if (errorOutputList.length > 0) {
            this.emit('error', new gutil.PluginError(meta.name, '\n' + errorOutputList.join('\n\n') + '\n', {
                showStack: false
            }));
        }
        callback();
    }

    return through.obj(transformFunction, errorOutput);
};
